#ifndef _DIVINE_COMPRESSOR_HH_
#define _DIVINE_COMPRESSOR_HH_

/*!\file
 * Unit for compression of explicit states of system. Implemented by
 * class compressor_t.
 *
 * \author Jiri Barnat
 */

#include "system/state.hh"
#include "common/error.hh"

#define NO_COMPRESS 11
#define HUFFMAN_COMPRESS 12

namespace divine {

//!Compression implementation class
class compressor_t {
private:
  char * compress_buffer;
  size_int_t compress_buffer_size;
public:
  //!A constructor
  compressor_t()
  {
   compress_buffer_size = 4096;
   compress_buffer = new char[compress_buffer_size];
  }
  
  //!A destructor
  ~compressor_t()
  {
   delete [] compress_buffer;
  }

  //! Compress a state according current compression method. 
  /*! Returns where the state is compressed and how long the compressed
   *  representation is. */
  bool compress(state_t, char *& pointer, int& size);
  
  //! Compress a state according current compression method, but returned pointer points to an internal buffer of compressor_t
  /*! Works the same way as compressor_t::compress(), but compressed state
   *  is stored internally in compressor_t, thus it does not need to be
   *  deleted */
  bool compress_without_alloc(state_t, char *& pointer, int& size);
  

  //! Decompress the state using current compression method
  /*! Decompress the state from the given pointer and size using current
   *  compression method. */
  bool decompress(state_t&, char *pointer, int size);
  
  //! Initializes compressor instance.
  /*! Requires identification of
   *  compression method, extra space that should be allocated at each
   *  compressed state (appendix) and pointer to explicit_system_t. */
  bool init(int method, int appendix_size);

  //! Clears data structures initialized inside compressor
  /*! Dealocates data that are created during initialization of compressor. This
   *  is needed especially for clearing the arena of compressor. */
  void clear();

private:
  int method_id;
  int oversize;
};

};

#endif


